//  AllStarLink WaterFall (allfall)
//  Willem A Scheuder
//  willem@prinmath.com
//
//  Version     Date       Remarks
//    1.0   Apr 23, 2021   Initial release
//    1.1   Apr 25, 2021   Add back up/advance
//    1.2   May  1, 2021   Add database and TCP client
//    1.2.1 May  7, 2021   Add TCP connection count
//    1.2.2 May 10, 2021   Improve Push sync
//    1.2.3 Nov 30, 2021   Fix sync bug
//    1.2.4 Dec 10, 2021   Less agressive locking in Poll thread
//
//  This code is released under the GNU Public License Version 2.

#define VER "1.2.4"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <pthread.h>
#include <time.h>
#include <fcntl.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>

#define Zero   0x00
#define Mix    0x10
#define Voter  0x20
#define Voted  0x30
#define NoRSSI 0xF0

//  For Windows use PDCurses
#ifdef __CYGWIN__
#include "pdcurses.h"
#else
#include <ncurses.h>
#endif
//  CYGWIN predefines optarg
#ifndef __CYGWIN__
extern char *optarg;
extern int  optind,opterr;
#endif

#ifdef MySQL
#include <mysql.h>
static MYSQL* mydb=0;
#else
const int mydb=0;
#endif

//  Help text
char* Help =
   "  ^    Scroll up\n"
   "  v    Scroll down\n"
   "  <    Back up\n"
   "  >    Advance\n"
   "  -    Back up page\n"
   "  +    Advance page\n"
   "  0    Real time\n"
   "  r    Reverse direction\n"
   "  h    Help\n"
   "  q    Quit program\n";

//  Max length of history
#define nsec 3600

// Client structure holds RSSI readings for a voter
typedef struct
{
   int num;            // Number of values
   int max;            // Max values
   int vote;           // Voting
   int node;           // Node number
   int id;             // Database id
   char* name;         // Client name
   uint8_t rssi[nsec]; // Buffer of replies
   pthread_mutex_t lock; //  Mutex to lock client for data updates
} Client;

// Stream structure holds Manager connection
typedef struct
{
   int sock;           // Socket
   int len;            // Data received
   char line[8192];    // Buffer of replies
   int nclient;        // Number of clients
   Client** client;    // List of clients
   pthread_t thread;   // Thread
   pthread_mutex_t lock; //  Mutex to lock stream for client addition
} Stream;

//  Ring buffer time
typedef struct
{
   time_t t;   // Current time
   int    now; // Ring buffer pointer
} Time;

int help=0;    // Show help
int sps=10;    // Samples per second
int r2l=1;     // Right to left
int wid=0;     // Window width
int hgt=0;     // Window height
int top=0;     // Top entry in display
int nwid=6;    // Name width
int rwid=0;    // RSSI width
int show=0;    // Update display
int run=1;     // Continue running
int dt=0;      // Time offset
int nstream=0; // Number of streams
int headless=0;          // Run as a daemon
int port=0;              // TCP port to listen on
int sock0=0;             // TCP port for listening
int npush=0;             // Number of TCP push threads
char ActionID[128];      // Action ID
pthread_t avg;           // Averaging thread
Stream** stream=NULL;    // Stream array
pthread_mutex_t dblock;  // Mutex lock for database
Time            upnow;   // Time and now pointer of update
pthread_rwlock_t uplock; // RW lock for update

//
//  Print error and exit
//
void Fatal(const char* format , ...)
{
   va_list args;
   //  wid>0 indicates curses active
   if (wid>0) endwin();
   //  Print error message
   va_start(args,format);
   vfprintf(stderr,format,args);
   va_end(args);
   //  Error exit
   exit(1);
}

//
//  Set update variable
//
static void upset(time_t t,int now)
{
   if (pthread_rwlock_wrlock(&uplock)) Fatal("Update write lock failed\n");
   upnow.t   = t;
   upnow.now = now;
   if (pthread_rwlock_unlock(&uplock)) Fatal("Update write unlock failed\n");
}

//
//  Get update variable
//
static Time upget(int off)
{
   if (pthread_rwlock_rdlock(&uplock)) Fatal("Update read lock failed\n");
   Time now = upnow;
   if (pthread_rwlock_unlock(&uplock)) Fatal("Update read unlock failed\n");
   //  Adjust for offset
   if (off)
   {
      now.t   -= off;
      now.now = (now.now+off)%nsec;
   }
   return now;
}

//
//  Allocate memory for string and copy
//
char* newstr(char* str0)
{
   int len = strlen(str0)+1;
   char* str = malloc(len);
   if (!str) Fatal("Cannot newstr length %d\n",len);
   strcpy(str,str0);
   return str;
}

//
//  Scroll up and down
//  dir=0 is resizing the screen
//
void Scroll(int dir)
{
   if (headless) return;
   //  Count rows
   int n=0;
   for (int j=0;j<nstream;j++)
      n += stream[j]->nclient;
   //  Everything fits
   if (n+2<hgt)
      top = 0;
   // Scroll up
   else if (dir<0 && top>0)
      top--;
   // Scroll down
   else if (dir>0 && top+hgt-2<n)
      top++;
   else if (dir==0 && top+hgt-2>=n)
      top = n-hgt-2;
}

//
//  Set window size
//
void Resize()
{
   if (headless) return;
   getmaxyx(stdscr,hgt,wid);
   Scroll(0);
   rwid = wid - nwid;
   if (rwid>nsec) rwid = nsec;
   if (dt>=nsec-rwid) dt = nsec-rwid-2;
}

//
//  Send string
//
void writestr(Stream* stream,char* str)
{
   int l = strlen(str);
   if (write(stream->sock,str,l)!=l) Fatal("Write error %s\n",str);
}

//
//  Read character
int readchar(Stream* stream)
{
   unsigned char ch;
   if (read(stream->sock,&ch,1)!=1) Fatal("Read error\n");
   return ch;
}

//
//  Read line
//
int readline(Stream* stream)
{
   int nl=0;
   stream->len = 0;
   while (!nl)
   {
      int ch = readchar(stream);
      //  Return on newline
      if (ch == '\n')
      {
         stream->line[stream->len] = 0;
         nl = 1;
      }
      //  Store all except \r
      else if (ch != '\r')
         stream->line[stream->len++] = ch;
   }
   return stream->len;
}

//
//  Find matching client
//
Client* matchclient(Stream* stream,int node,char* name)
{
   for (int k=0;k<stream->nclient;k++)
      if (stream->client[k]->node==node && !strcmp(stream->client[k]->name,name))
         return stream->client[k];
   return NULL;
}

//
//  Allocate and initialize client
//
Client* NewClient(int node,char* name)
{
   //  Name width
   int l = strlen(name);
   if (nwid<l) nwid = l;
   //  Allocate memory
   Client* client = malloc(sizeof(Client));
   if (!client) Fatal("Cannot malloc node %d client %s\n",node,name);
   //  Initialize
   client->num = 0;                  // Number of values
   client->max = 0;                  // Sum of values
   client->id = -1;                  // Database ID
   client->vote = -1;                // Voting
   client->node = node;              // Node
   client->name = newstr(name);      // Client name
   memset(client->rssi,NoRSSI,nsec); // Initialize rssi
   if (pthread_mutex_init(&client->lock,NULL)) Fatal("Cannot init client lock\n");
#ifdef MySQL
   //  Check if client is defined in database, insert if new
   if (pthread_mutex_lock(&dblock)) Fatal("DB lock failed\n");
   if (mydb)
   {
      char sql[1024];
      sprintf(sql,"SELECT id FROM client WHERE node=%d AND name='%s'",node,name);
      if (mysql_query(mydb,sql)) Fatal("Client Select Fail %s\n",sql);
      MYSQL_RES* res = mysql_store_result(mydb);
      if (!res)  Fatal("Client Select Error %s\n",sql);
      //  Found match
      if (mysql_num_rows(res))
      {
         MYSQL_ROW row = mysql_fetch_row(res);
         client->id = atoi(row[0]);
         mysql_free_result(res);
      }
      else
      {
         mysql_free_result(res);
         sprintf(sql,"INSERT INTO client(node,name) VALUES(%d,'%s')",node,name);
         if (mysql_query(mydb,sql)) Fatal("Client Insert Fail %s\n",sql);
         client->id = mysql_insert_id(mydb);
      }
   }
   if (pthread_mutex_unlock(&dblock)) Fatal("DB lock failed\n");
#endif
   //  Return new pointer
   return client;
}

//
//  Poll stream
//
void* Poll(Stream* stream)
{
   while (run)
   {
      // Send request
      writestr(stream,"ACTION: VoterStatus\r\n");
      writestr(stream,ActionID);
      writestr(stream,"\r\n");
      // Read replies until empty line
      int node=-1;
      Client* client=NULL;
      Client* vote=NULL;
      while (readline(stream))
      {
         //  Locate ': ' and split line on that
         char* ch = strstr(stream->line,": ");
         if (!ch) continue;
         *ch = 0;
         ch+=2;
         //  Remove  Master or/or ActiveMaster
         char* cx;
         if ((cx = strstr(ch," Master ActiveMaster")) && !*cx+21)
            *cx = 0;
         else if ((cx = strstr(ch," ActiveMaster")) && !*cx+14)
            *cx = 0;
         else if ((cx = strstr(ch," Master")) && !*cx+8)
            *cx = 0;
         //  Node number
         if (!strcmp(stream->line,"Node"))
         {
            node = atoi(ch);
            client = vote = NULL;
         }
         //  Voted
         else if (!strcmp(stream->line,"Voted"))
            vote = matchclient(stream,node,ch);
         //  Client (add if new)
         else if (!strcmp(stream->line,"Client"))
         {
            //  Matched client
            client = matchclient(stream,node,ch);
            //  Add new client to stream
            if (!client)
            {
               if (pthread_mutex_lock(&stream->lock)) Fatal("Poll stream lock failed\n");
               stream->client = (Client**)realloc(stream->client,(stream->nclient+1)*sizeof(Client));
               stream->client[stream->nclient++] = client = NewClient(node,ch);
               if (pthread_mutex_unlock(&stream->lock)) Fatal("Poll stream unlock failed\n");
               Resize();
            }
         }
         //  RSSI (must have valid client)
         else if (!strcmp(stream->line,"RSSI") && client)
         {
            int rssi = atoi(ch);
            if (pthread_mutex_lock(&client->lock)) Fatal("Poll client lock failed\n");
            client->num++;
            if (client->max < rssi) client->max = rssi;
            if (vote && client->vote<0) client->vote = 0;
            if (client==vote) client->vote++;
            if (pthread_mutex_unlock(&client->lock)) Fatal("Poll client unlock failed\n");
         }
      }
      //  Pause until next sample
      usleep(1000000/sps);
   }
   return NULL;
}

//
//  Push string
//
int PushString(int sock,char* str)
{
   int l = strlen(str);
   return write(sock,str,l)-l;
}

//
//  Push data strings and increment now
//
int PushData(int sock,Time* now)
{
   if (sock<0) return 0;
   //  Increment now
   now->t++;
   now->now = (now->now+nsec-1)%nsec;
   //  Time offset from 2020-01-01 00:00:00
   int t = now->t-1577862000;
   char timestr[32];
   sprintf(timestr,"%d",t);
   if (PushString(sock,timestr)) return 1;
   //  Loop over streams and clients
   int n = now->now;
   for (int j=0;j<nstream;j++)
      for (int k=0;k<stream[j]->nclient;k++)
      {
         //  Get RSSI
         char buf[8];
         char* str = buf;
         *str++ = '\t';
         uint8_t ch = stream[j]->client[k]->rssi[n];
         //  No RSSI
         if (ch==NoRSSI)
            *str++ = ' ';
         //  Nothing received
         else if (ch==Zero)
         {
            *str++ = '-';
         }
         // 0-9 or #
         else
         {
            char tp = (ch>>4);
            if (tp==1)
               *str++ = 'M';
            else if (tp==2)
               *str++ = 'v';
            else if (tp==3)
               *str++ = 'V';
            ch &= 0xF;
            if (ch>9)
               *str++ = '#';
            else
               *str++ = ch+'0';
         }
         *str++ = 0;
         if (PushString(sock,buf)) return 1;
      }
   //  End of data
   return PushString(sock,"\n");
}

//
//  Push client strings
//
int PushClient(int sock)
{
   if (sock<0) return 0;
   if (PushString(sock,"0")) return 1;
   //  Loop over streams and clients
   for (int j=0;j<nstream;j++)
      for (int k=0;k<stream[j]->nclient;k++)
      {
         if (PushString(sock,"\t")) return 1;
         // Push client name
         if (PushString(sock,stream[j]->client[k]->name)) return 1;
      }
   return PushString(sock,"\n");
}

//
//  Read integer from line (no whitespace allowed other than \r)
//
int readint(int sock)
{
   char ch;
   int n=0;
   while(read(sock,&ch,1))
   {
      if ('0'<=ch && ch<='9')
        n = 10*n + ch-'0';
      else if (ch=='\n')
         return n;
      else if (ch!='\r')
         return -1;
   }
   return -1;
}

//
//  Push updates
//    Close thread on any failures
//
void* Push(int* sock0)
{
   //  Accept connection
   struct sockaddr_in sin;
   socklen_t l = sizeof(sin);
   int sock = accept(*sock0,(struct sockaddr*)&sin,&l);
   if (sock<0) return NULL;
   npush++;
   // Read how much history the client wants
   int n = readint(sock);
   if (n<0)
   {
      close(sock);
      sock = -1;
   }
   else if (n>nsec-2)
      n = nsec-2;
   // Push client list
   if (PushClient(sock))
   {
      close(sock);
      sock = -1;
   }
   // Update loop
   Time t0 = upget(n);
   while (sock>=0 && run)
   {
      Time now = upget(0);
      // Push updates
      while (sock>=0 && now.t>t0.t)
      {
         if (PushData(sock,&t0))
         {
            close(sock);
            sock = -1;
         }
      }
      //  Sleep 10ms
      usleep(10000);
   }
   npush--;
   return NULL;
}

//
//  Create and initialize new stream
//
Stream* NewStream(int id,char* host,int port,char* user,char* pass)
{
   // Create socket
   int sock = socket(AF_INET,SOCK_STREAM,0);
   if (sock<0) Fatal("Failed to create socket for %s:%d\n",host,port);

   //  Resolve hostname
   struct hostent* he = gethostbyname(host);
   if (!he) Fatal("Cannot resolve host name %s\n",host);
   if (port<1 || port>65535) Fatal("Invalid port %s:%d\n",host,port);

   // Set address
   struct sockaddr_in addr;
   bzero(&addr,sizeof(addr));
   addr.sin_family = AF_INET;
   memcpy(&addr.sin_addr.s_addr,he->h_addr_list[0],4);
   addr.sin_port = htons(port);

   // Connect to server
   if (connect(sock,(struct sockaddr*)&addr,sizeof(addr))) Fatal("Failed to connect to %s:%d\n",host,port);

   //  Allocate structure
   Stream* stream = malloc(sizeof(Stream));
   if (!stream) Fatal("Cannot allocate stream for %s:%d\n",host,port);
   stream->sock  = sock;
   stream->len   = 0;
   stream->nclient = 0;
   stream->client  = NULL;

   // Login
   writestr(stream,"ACTION: LOGIN\r\n");
   writestr(stream,"USERNAME: ");writestr(stream,user);writestr(stream,"\r\n");
   writestr(stream,"SECRET: ");writestr(stream,pass);writestr(stream,"\r\n");
   writestr(stream,"EVENTS: 0\r\n");
   writestr(stream,ActionID);
   writestr(stream,"\r\n");

   // Check for valid login
   int login=0;
   while (readline(stream))
   {
      if (!strstr(stream->line,"Response: Success")) login++;
   }
   if (!login) Fatal("Login failed for %s:%d\n",host,port);

   //  Start read thread
   if (pthread_create(&stream->thread,NULL,(void*)Poll,stream)) Fatal("Cannot start thread for %s:%d\n",host,port);
   //  Create lock for stream data
   if (pthread_mutex_init(&stream->lock,NULL)) Fatal("Cannot create lock for %s:%d\n",host,port);

   //  Return new stream
   return stream;
}

//
//  Read configuration file
//
void ReadConfig(char* file)
{
   //  Action ID
   sprintf(ActionID,"ActionID: %d\r\n",getpid());
   //  Open first file in the list that is readable
   FILE* f = fopen(file,"r");
   if (!f) Fatal("Cannot open file %s",file);
   //  Read first 3 characters to check BOM
   unsigned char magic[3];
   if (fread(magic,1,3,f)==3 && magic[0]==0xEF && magic[1]==0xBB && magic[2]==0xBF)
      fprintf(stderr,"WARNING: UTF-8 file treated as ASCII\n");
   else
      rewind(f);

   //  Read line
   char line[8192];
   while (fgets(line,sizeof(line),f) != NULL )
   {
      char host[256];
      char user[256];
      char pass[256];
      //  Skip comments
      if (line[0] == '#') continue;
      //  Trim trailing whitespace
      int l=strlen(line)-1;
      while (l>0 && isspace((int)line[l-1])) l--;
      line[l] = 0;
      //  Skip blank lines
      if (l==0) continue;
      //  Get data
      if (sscanf(line,"%255s %255s %255s",host,user,pass)!=3) Fatal("Invalid data %s\n",line);
      //  Get port
      int  port = 5038;
      char* ch = strstr(host,":");
      if (ch)
      {
         *ch++ = 0;
         port = atoi(ch);
      }
      //  Allocate stream pointer memory
      stream = (Stream**)realloc(stream,(nstream+1)*sizeof(Stream*));
      //  Add to array
      stream[nstream] = NewStream(nstream,host,port,user,pass);
      //  Increment stream count
      nstream++;
   }
   fclose (f);
   if (!nstream) Fatal("No streams in %s\n",file);
}

//
//  Set color based on high nibble
//
void SetColor(char ch)
{
   ch = (ch>>4);
   if (0<ch && ch<4)
      attron(COLOR_PAIR(ch+1));
   else
      attron(COLOR_PAIR(1));
}

//
//  Draw single RSSI
//
void DrawRSSI(Client* client,int l)
{
   uint8_t ch = client->rssi[l];
   SetColor(ch);
   //  No RSSI
   if (ch==NoRSSI)
   {
      addch(' ');
   }
   //  No RSSI
   else if (ch==Zero)
   {
      addch('-');
   }
   // 0-9 or #
   else
   {
      ch &= 0xF;
      attron(A_BOLD);
      if (ch>9)
         addch('#');
      else
         addch(ch+'0');
      attroff(A_BOLD);
   }
   attron(COLOR_PAIR(1));
}

//
//  Draw client
//
void DrawClient(Client* client,Time now)
{
   //  Time to show
   int n = now.now;
   //  Print client name
   char* ch = client->name;
   SetColor(client->rssi[n]);
   for (int l=0;l<nwid;l++)
      addch(*ch?*ch++:'.');
   attron(COLOR_PAIR(1));
   //  Draw RSSI
   if (r2l)
      for (int l=rwid-1;l>=0;l--)
         DrawRSSI(client,(l+n)%nsec);
   else
      for (int l=0;l<rwid;l++)
         DrawRSSI(client,(l+n)%nsec);
}

//
//  Print history header
//
int PrintHeader(Time now)
{
   //  Header line
   time_t t = now.t-dt;
   struct tm* l = localtime(&t);
   if (dt>0) printw("%d ",dt);
   printw("%4d-%.2d-%.2d %.2d:%.2d:%.2d",l->tm_year+1900,l->tm_mon+1,l->tm_mday,l->tm_hour,l->tm_min,l->tm_sec);
   attron(A_BOLD);
   attron(COLOR_PAIR(2));
   printw(" Mix");
   attron(COLOR_PAIR(3));
   printw(" Voting");
   attroff(A_BOLD);
   attron(COLOR_PAIR(4));
   printw(" Voted");
   attron(COLOR_PAIR(1));
   printw(" TCP %d\n",npush);

   //  Client name
   attron(A_BOLD);
   printw("Client");
   for (int l=6;l<nwid;l++)
      addch(' ');
   //  Header right to left
   if (r2l)
      for (int l=rwid-1;l>=0;l--)
      {
         if (l%10==0)
            addch('0');
         else if (l>10 && (l-1)%10==0)
            addch('0'+(l%100)/10);
         else if (l>100 && (l-2)%10==0)
            addch('0'+l/100);
         else
            addch(' ');
      }
   //  Header left to right
   else
      for (int l=0;l<rwid;l++)
      {
         if (l%10==0)
            addch('0');
         else if (l>8 && (l+1)%10==0)
            addch('0'+((l+1)%100)/10);
         else if (l>97 && (l+2)%10==0)
            addch('0'+(l+2)/100);
         else
            addch(' ');
      }
   //  Unbold
   attroff(A_BOLD);
   return 2;
}

//
//  Display
//
void Display()
{
   show = 0;
   Time now =upget(dt);
   //  No output in daemon mode
   if (headless) return;
   //  Clear
   erase();
   attron(COLOR_PAIR(1));
   //  Help
   if (help)
   {
      attron(A_BOLD);
      printw(" Key   Function\n");
      attroff(A_BOLD);
      printw(Help);
   }
   //  RSSI
   else
   {
      //  Header
      int i=PrintHeader(now);
      //  Draw Client table
      int l=0;
      for (int j=0;j<nstream;j++)
         for (int k=0;k<stream[j]->nclient;k++)
            if (top<++l && i<hgt)
            {
               move(i++,0);
               DrawClient(stream[j]->client[k],now);
            }
   }
   refresh();
}

//
//  Average RSSI and advance now pointer
//    This thread also pushes data to the database
//    and accepts connections from TCP clients
//
void* Average()
{
   //  Allocate memory for update string
   //  Listening port
   if (port)
   {
      //  Prevent write errors from crashing threads
      signal(SIGPIPE,SIG_IGN);
      //  Open TCP socket
      sock0 = socket(AF_INET,SOCK_STREAM,0);
      if (sock0<0) Fatal("Cannot open TCP socket on port %d\n",port);
      //  Bind to port
      struct sockaddr_in addr;
      bzero(&addr, sizeof(addr));
      addr.sin_family = AF_INET;
      addr.sin_addr.s_addr = htonl(INADDR_ANY);
      addr.sin_port = htons(port);
      if (bind(sock0,(struct sockaddr*)&addr,sizeof(addr))<0) Fatal("Cannot bind socket to TCP port %d\n",port);
      // Listen for connections (up to 8 at a time)
      if (listen(sock0,8)<0) Fatal("Cannot listen to TCP port %d\n",port);
   }
#ifdef MySQL
   //  MySQL connection
   MYSQL_STMT* stmt=0;
   MYSQL_BIND  bind[3];
   MYSQL_TIME  mytime;
   int16_t id;
   int16_t rssi;
   if (mydb)
   {
      //  Lock DB
      if (pthread_mutex_lock(&dblock)) Fatal("DB lock failed\n");
      //  Prepare INSERT
      stmt = mysql_stmt_init(mydb);
      char* sql = "INSERT INTO rssi(id,time,rssi) VALUES(?,?,?)";
      if (mysql_stmt_prepare(stmt,sql,strlen(sql))) Fatal("RSSI Prep Failed\n");
      //  Bind data for INSERT
      memset(bind, 0, sizeof(bind));
      bind[0].buffer_type = MYSQL_TYPE_SHORT;
      bind[0].buffer = &id;
      bind[0].buffer_length = sizeof(id);
      bind[1].buffer_type = MYSQL_TYPE_DATETIME;
      bind[1].buffer = &mytime;
      bind[1].buffer_length = sizeof(mytime);
      bind[2].buffer_type = MYSQL_TYPE_SHORT;
      bind[2].buffer = &rssi;
      bind[2].buffer_length = sizeof(rssi);
      if (mysql_stmt_bind_param(stmt,bind)) Fatal("RSSI Bind Failed\n");
      //  Unlock DB
      if (pthread_mutex_unlock(&dblock)) Fatal("DB unlock failed\n");
   }
#endif
   //  RW lock of now and time
   time_t t0 = time(NULL);
   int    now=0;
   if (pthread_rwlock_init(&uplock,NULL)) Fatal("Cannot init uplock\n");
   upset(t0,now);
   while (run)
   {
      time_t t = time(NULL);
      //  Catch up to wall time
      while (t0<t)
      {
         //  Update now
         t0++;
         now = (now+nsec-1)%nsec;
#ifdef MySQL
         //  Time for database
         struct tm* lt = localtime(&t0);
         mytime.year   = lt->tm_year+1900;
         mytime.month  = lt->tm_mon+1;
         mytime.day    = lt->tm_mday;
         mytime.hour   = lt->tm_hour;
         mytime.minute = lt->tm_min;
         mytime.second = lt->tm_sec;
         mytime.neg    = 0;
         mytime.second_part = 0;
#endif
         //  Increase dt if positive and moving
         if (0<dt && dt<nsec-rwid) dt++;
         //
         //  Average all streams
         //
         for (int j=0;j<nstream;j++)
         {
            //  Lock stream (don't need dblock because stream lock prevents new clients on this stream)
            if (pthread_mutex_lock(&stream[j]->lock)) Fatal("Average stream lock failed %d\n",j);
            //  Average clients
            for (int k=0;k<stream[j]->nclient;k++)
            {
#ifdef MySQL
               //  ID=0 means no data
               id = 0;
#endif
               //  Grab client
               Client* client = stream[j]->client[k];
               if (pthread_mutex_lock(&client->lock)) Fatal("Average client lock failed %d\n",j);
               //  No RSSI
               if (client->num==0)
                  client->rssi[now] = NoRSSI;
               //  All readings are zero
               else if (client->max==0)
                  client->rssi[now] = Zero;
               //  Show RSSI and voting status
               else
               {
                  int r10 = client->max/25.5;
                  if (client->vote<0)
                     r10 += Mix;
                  else if (client->vote==0)
                     r10 += Voter;
                  else
                     r10 += Voted;
                  client->rssi[now] = r10;
#ifdef MySQL
                  //  Data to store to database
                  id   = client->id;
                  rssi = client->max;
                  if (client->vote>0)
                     rssi += 0x200;
                  else if (client->vote==0)
                     rssi += 0x100;
#endif
               }
               //  Reset client
               client->num = 0;
               client->max = 0;
               //  Once voted never go back
               if (client->vote>0) client->vote = 0;
               //  Unlock client
               if (pthread_mutex_unlock(&client->lock)) Fatal("Average client unlock failed %d\n",j);
#ifdef MySQL
               // Push reading to database
               if (mydb && id>0 && mysql_stmt_execute(stmt)) Fatal("RSSI Insert Failed\n");
#endif
            }
            //  Unlock this stream from having new clients added
            if (pthread_mutex_unlock(&stream[j]->lock)) Fatal("Average stream unlock failed %d\n",j);
         }
         // Check for new connections
         if (port)
         {
            //  Time out immediately
            struct timeval timer = {0,0};
            //  TCP socket only
            fd_set set;
            FD_ZERO(&set);
            FD_SET(sock0,&set);
            //  Check for connection
            if (select(sock0+1,&set,NULL,NULL,&timer) && FD_ISSET(sock0,&set))
            {
               //  Launch detached thread
               pthread_t thid;
               if (pthread_create(&thid,NULL,(void*)Push,&sock0)) Fatal("Cannot launch Push thread\n");
               if (pthread_detach(thid)) Fatal("Cannot detach Push thread\n");
            }
         }
         // Update display and remember time
         upset(t,now);
         show = 1;
      }
      //  Sleep 1ms
      usleep(1000);
   }
   return NULL;
}

//
//  Initialize curses
//
void InitCurses()
{
   // Start curses with raw input
   initscr();
   nonl(); cbreak(); noecho(); nodelay(stdscr,TRUE); keypad(stdscr,TRUE); curs_set(0);
   // Set window size
   Resize();
   //  Set colors
   if (!has_colors()) Fatal("No color support\n");
   start_color();
   init_color(COLOR_WHITE,1000,1000,1000);
   init_pair(1,COLOR_WHITE,COLOR_BLACK);
   init_pair(2,COLOR_CYAN,COLOR_BLACK);
   init_pair(3,COLOR_GREEN,COLOR_BLACK);
   init_pair(4,COLOR_YELLOW,COLOR_BLACK);
   init_pair(5,COLOR_RED,COLOR_BLACK);
   bkgd(COLOR_PAIR(1));
}

//
//  Initialize database
//
void DBinit(char* arg)
{
#ifndef MySQL
   Fatal("Database support not selected at compile time\n");
#else
   if (!(mydb = mysql_init(NULL))) Fatal("Cannot initialize database\n");
   //  Parse database connection string
   //  server:database/user:pass
   char* server = newstr(arg);
   if (!server) Fatal("Cannot allocate working string for database name\n");
   strcpy(server,arg);
   //  Split database from user
   char* user = strstr(server,"/");
   if (!user) Fatal("Missing /user:pass in database name %s\n",arg);
   *user++ = 0;
   //  : in server separates server and database
   //  absent : server is assumed to be localhost
   char* database = strstr(server,":");
   if (database)
      *database++ = 0;
   else
   {
      database = server;
      server = "localhost";
   }
   //  : in server separates user and password
   //  absent : password is null
   char* pass = strstr(user,":");
   if (pass) *pass++ = 0;
   //  Connect
   int port = 3306;
   if (!mysql_real_connect(mydb,server,user,pass,database,port,NULL,CLIENT_COMPRESS)) Fatal("Failed to connect to database %s on %s as %s password %s\n",database,server,user,pass);
   //  Create mutex lock for databse
   if (pthread_mutex_init(&dblock,NULL)) Fatal("Cannot create lock for database\n");
   free(server);
#endif
}

//
//  Main
//
int main(int argc,char* argv[])
{
   //
   //  Process command line arguments
   //
   int ch;
   while ((ch = getopt(argc,argv,"vhrs:d:p:D")) != EOF)
   {
       //  Run as daemon
       if (ch == 'D')
          headless = 1;
       //  Reverse direction
       else if (ch == 'r')
          r2l = 0;
       //  Samples per second
       else if (ch == 's')
       {
          sps = atoi(optarg);
          if (sps<1 || sps>20) Fatal("Invalid samples per second %d\n",sps);
       }
       //  TCP port
       else if (ch == 'p')
       {
          port = atoi(optarg);
          if (port<1 || port>65535) Fatal("Invalid port %d\n",port);
       }
       //  Connect to database
       else if (ch == 'd')
          DBinit(optarg);
       //  Help
       else if (ch == 'h')
          Fatal("Usage: allfall [-rsvhD] [-p port] [-d host:db/user:pass] file\n"
                "  -r  Scroll RSSI left to right\n"
                "  -s  samples per second\n"
                "  -D  run headless\n"
                "  -v  version\n"
                "  -h  help\n"
                " Each line in file is host:port username password\n"
                " # in column 1 starts a comment\n\n"
                " Key  Function\n%s",Help);
       //  Version
       else if (ch == 'v')
          Fatal("allfall version " VER "\n");
       else
          Fatal("Unknown option %c\n",ch);
   }
   if (headless && !mydb && !port) Fatal("Running headless without a database or TCP is pointless\n");
   argc -= optind;
   argv += optind;
   if (argc!=1) Fatal("Usage: allfall [-rsvh] file\n");
   //  Read data (also fires of threads)
   ReadConfig(argv[0]);
   //  Initialize curses
   if (!headless) InitCurses();
   //  Averaging thread
   if (pthread_create(&avg,NULL,(void*)Average,NULL)) Fatal("Cannot start averaging thread\n");
   //  Main loop
   while(run)
   {
      int ch = getch();
      //  Quit
      if (ch=='q')
         run = 0;
      //  Window resized
      else if (ch==KEY_RESIZE)
      {
         Resize();
         Display();
      }
      //  Back up
      else if (ch==KEY_LEFT)
      {
         if (dt<nsec-rwid) dt++;
         Display();
      }
      //  Advance
      else if (ch==KEY_RIGHT)
      {
         if (dt>0) dt--;
         Display();
      }
      //  Back up page
      else if (ch=='-')
      {
         dt += rwid;
         if (dt>=nsec-rwid) dt=nsec-rwid-2;
         Display();
      }
      //  Advance page
      else if (ch=='+')
      {
         dt -= rwid;
         if (dt<0) dt = 0;
         Display();
      }
      //  Make current
      else if (ch=='0')
      {
         dt = 0;
         Display();
      }
      //  Scroll down
      else if (ch==KEY_DOWN)
      {
         Scroll(+1);
         Display();
      }
      //  Scroll down
      else if (ch==KEY_NPAGE)
      {
         Scroll(+hgt);
         Display();
      }
      //  Scroll up
      else if (ch==KEY_UP)
      {
         Scroll(-1);
         Display();
      }
      //  Scroll up
      else if (ch==KEY_PPAGE)
      {
         Scroll(-hgt);
         Display();
      }
      //  Reverse RSSI display direction
      else if (ch=='r')
      {
         r2l = !r2l;
         Display();
      }
      //  Show help
      else if (ch=='h')
      {
         help = !help;
         show = 1;
      }
      //  Update display
      else if (show)
         Display();
      //  Sleep 10ms
      usleep(10000);
   }
   endwin();
#ifdef MySQL
   //  Close database if opened
   if (mydb) mysql_close(mydb);
#endif
   return 0;
}
